@@ -35,6 +35,9 @@ from base64 import b64encode
 from time import time
 from datetime import datetime, date, timezone
 from decimal import Decimal
+from uuid import UUID
+
+import urllib3
 from urllib3 import connection_from_url
 from urllib3.connection import HTTPConnection
 from urllib3.exceptions import (
@@ -46,11 +49,14 @@ from urllib3.exceptions import (
     SSLError,
 )
 from urllib3.util.retry import Retry
+
+from crate.client._pep440 import Version
 from crate.client.exceptions import (
     ConnectionError,
     BlobLocationNotFoundException,
     DigestNotFoundException,
     ProgrammingError,
+    IntegrityError,
 )
 
 
@@ -86,7 +92,7 @@ class CrateJsonEncoder(json.JSONEncoder):
     epoch_naive = datetime(1970, 1, 1)
 
     def default(self, o):
-        if isinstance(o, Decimal):
+        if isinstance(o, (Decimal, UUID)):
             return str(o)
         if isinstance(o, datetime):
             if o.tzinfo is not None:
@@ -186,6 +192,18 @@ def _ex_to_message(ex):
 
 
 def _raise_for_status(response):
+    """
+    Properly raise `IntegrityError` exceptions for CrateDB's `DuplicateKeyException` errors.
+    """
+    try:
+        return _raise_for_status_real(response)
+    except ProgrammingError as ex:
+        if "DuplicateKeyException" in ex.message:
+            raise IntegrityError(ex.message, error_trace=ex.error_trace) from ex
+        raise
+
+
+def _raise_for_status_real(response):
     """ make sure that only crate.exceptions are raised that are defined in
     the DB-API specification """
     message = ''
@@ -253,10 +271,11 @@ def _pool_kw_args(verify_ssl_cert, ca_cert, client_cert, client_key,
         'cert_reqs': ssl.CERT_REQUIRED if verify_ssl_cert else ssl.CERT_NONE,
         'cert_file': client_cert,
         'key_file': client_key,
-        'timeout': timeout,
     }
+    if timeout is not None:
+        kw['timeout'] = float(timeout)
     if pool_size is not None:
-        kw['maxsize'] = pool_size
+        kw['maxsize'] = int(pool_size)
     return kw
 
 
@@ -271,6 +290,19 @@ def _remove_certs_for_non_https(server, kwargs):
     return kwargs
 
 
+def _update_pool_kwargs_for_ssl_minimum_version(server, kwargs):
+    """
+    On urllib3 v2, re-add support for TLS 1.0 and TLS 1.1.
+
+    https://urllib3.readthedocs.io/en/latest/v2-migration-guide.html#https-requires-tls-1-2
+    """
+    if Version(urllib3.__version__) >= Version("2"):
+        from urllib3.util import parse_url
+        scheme, _, host, port, *_ = parse_url(server)
+        if scheme == "https":
+            kwargs["ssl_minimum_version"] = ssl.TLSVersion.MINIMUM_SUPPORTED
+
+
 def _create_sql_payload(stmt, args, bulk_args):
     if not isinstance(stmt, str):
         raise ValueError('stmt is not a string')
@@ -301,7 +333,7 @@ def _get_socket_opts(keepalive=True,
     # always use TCP keepalive
     opts = [(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)]
 
-    # hasattr check because some of the options depend on system capabilities
+    # hasattr check because some options depend on system capabilities
     # see https://docs.python.org/3/library/socket.html#socket.SOMAXCONN
     if hasattr(socket, 'TCP_KEEPIDLE') and tcp_keepidle is not None:
         opts.append((socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, tcp_keepidle))
@@ -337,6 +369,7 @@ class Client(object):
                  error_trace=False,
                  cert_file=None,
                  key_file=None,
+                 ssl_relax_minimum_version=False,
                  username=None,
                  password=None,
                  schema=None,
@@ -377,6 +410,7 @@ class Client(object):
             'socket_tcp_keepintvl': socket_tcp_keepintvl,
             'socket_tcp_keepcnt': socket_tcp_keepcnt,
         })
+        self.ssl_relax_minimum_version = ssl_relax_minimum_version
         self.backoff_factor = backoff_factor
         self.server_pool = {}
         self._update_server_pool(servers, **pool_kw)
@@ -397,6 +431,10 @@ class Client(object):
 
     def _create_server(self, server, **pool_kw):
         kwargs = _remove_certs_for_non_https(server, pool_kw)
+        # After updating to urllib3 v2, optionally retain support for TLS 1.0 and TLS 1.1,
+        # in order to support connectivity to older versions of CrateDB.
+        if self.ssl_relax_minimum_version:
+            _update_pool_kwargs_for_ssl_minimum_version(server, kwargs)
         self.server_pool[server] = Server(server, **kwargs)
 
     def _update_server_pool(self, servers, **pool_kw):
