@@ -35,17 +35,19 @@ from unittest.mock import patch, MagicMock
 from threading import Thread, Event
 from decimal import Decimal
 import datetime as dt
+
 import urllib3.exceptions
 from base64 import b64decode
 from urllib.parse import urlparse, parse_qs
-from setuptools.ssl_support import find_ca_bundle
+
+import uuid
+import certifi
 
 from .http import Client, CrateJsonEncoder, _get_socket_opts, _remove_certs_for_non_https
-from .exceptions import ConnectionError, ProgrammingError
-
+from .exceptions import ConnectionError, ProgrammingError, IntegrityError
 
 REQUEST = 'crate.client.http.Server.request'
-CA_CERT_PATH = find_ca_bundle()
+CA_CERT_PATH = certifi.where()
 
 
 def fake_request(response=None):
@@ -88,6 +90,17 @@ def bad_bulk_response():
     return r
 
 
+def duplicate_key_exception():
+    r = fake_response(409, 'Conflict')
+    r.data = json.dumps({
+        "error": {
+            "code": 4091,
+            "message": "DuplicateKeyException[A document with the same primary key exists already]"
+        }
+    }).encode()
+    return r
+
+
 def fail_sometimes(*args, **kwargs):
     if random.randint(1, 100) % 10 == 0:
         raise urllib3.exceptions.MaxRetryError(None, '/_sql', '')
@@ -287,6 +300,30 @@ class HttpClientTest(TestCase):
         )
         client.close()
 
+    @patch(REQUEST, autospec=True)
+    def test_uuid_serialization(self, request):
+        client = Client(servers="localhost:4200")
+        request.return_value = fake_response(200)
+
+        uid = uuid.uuid4()
+        client.sql('insert into my_table (str_col) values (?)', (uid,))
+
+        data = json.loads(request.call_args[1]['data'])
+        self.assertEqual(data['args'], [str(uid)])
+        client.close()
+
+    @patch(REQUEST, fake_request(duplicate_key_exception()))
+    def test_duplicate_key_error(self):
+        """
+        Verify that an `IntegrityError` is raised on duplicate key errors,
+        instead of the more general `ProgrammingError`.
+        """
+        client = Client(servers="localhost:4200")
+        with self.assertRaises(IntegrityError) as cm:
+            client.sql('INSERT INTO testdrive (foo) VALUES (42)')
+        self.assertEqual(cm.exception.message,
+                         "DuplicateKeyException[A document with the same primary key exists already]")
+
 
 @patch(REQUEST, fail_sometimes)
 class ThreadSafeHttpClientTest(TestCase):
